# -*- encoding:utf-8 -*-
"""
@author：lin.li
@filename：__init__.py
@time：2022/8/15  21:36
@file description: pubmed_download
"""
from .pubmed_download import DownloadPdf
from .pubmed_mapper import Article



# -*- coding: utf-8 -*-
"""
PubMed Mapper: A Python library that map PubMed XML to Python object
"""
import re
from datetime import date
from urllib.request import urlopen
from urllib.error import URLError, HTTPError
from lxml import etree


MONTHS = {
    'Jan': 1, 'Feb': 2, 'Mar': 3,
    'Apr': 4, 'May': 5, 'Jun': 6,
    'Jul': 7, 'Aug': 8, 'Sep': 9,
    'Sept': 9, 'Oct': 10, 'Nov': 11,
    'Dec': 12,
    'January': 1, 'February': 2, 'March': 3,
    'April': 4,  # 'May': 5,
    'June': 6, 'July': 7, 'August': 8, 'September': 9,
    'October': 10, 'November': 11, 'Novembre': 11, 'December': 12
}

SEASONS = {
    'Spring': 4, 'Summer': 7,
    'Fall': 10, 'Autumn': 10, 'Winter': 1
}


class PubmedMapperError(Exception):
    """PubmedMapper Error"""
    pass


def extract_first(data):
    if not (isinstance(data, list) and (len(data) >= 1)):
        return None
    return data[0]


def get_inner_html(element, strip=True):
    texts = []
    if element is None:
        return ""
    if element.text:
        texts.append(element.text)
    for child in element.getchildren():
        if child.text:
            texts.append(child.text)
        if child.tail:
            texts.append(child.tail)
    if element.tail:
        texts.append(element.tail)
    text = ''.join(texts)
    if strip:
        text = text.strip()
    return text


class ArticleId(object):
    def __init__(self, id_type, id_value):
        self.id_type = id_type
        self.id_value = id_value

    def __repr__(self):
        return '%s: %s' % (self.id_type, self.id_value)

    def to_dict(self):
        return {
            'id_type': self.id_type,
            'id_value': self.id_value
        }

    @classmethod
    def parse_element(cls, element):
        return cls(
            id_type=element.get('IdType'),
            id_value=element.text
        )


class Mesh(object):
    def __init__(self, id_type, id_value):
        self.id_type = id_type
        self.id_value = id_value

    def __repr__(self):
        return '%s: %s' % (self.id_type, self.id_value)

    def to_dict(self):
        return {
            'id_type': self.id_type,
            'id_value': self.id_value
        }

    @classmethod
    def parse_element(cls, element):
        return cls(
            id_type=element.get('UI'),
            id_value=element.text
        )


class Publication(object):
    def __init__(self, id_type, id_value):
        self.id_type = id_type
        self.id_value = id_value
    
    def __repr__(self):
        return '%s: %s' % (self.id_type, self.id_value)
    
    def to_dict(self):
        return {
            'id_type': self.id_type,
            'id_value': self.id_value
        }
    
    @classmethod
    def parse_element(cls, element):
        return cls(
            id_type=element.get('UI'),
            id_value=element.text
        )
    

class PubdateDefaults(object):
    """default year, month, day"""
    default_year = 1
    default_month = 1
    default_day = 1


class PubdateParserYearMonthDay(PubdateDefaults):
    """年份、月份、日期都有"""
    def __call__(self, element):
        year_text = extract_first(element.xpath('./Year/text()'))
        month_text = extract_first(element.xpath('./Month/text()'))
        day_text = extract_first(element.xpath('./Day/text()'))
        if not (year_text and month_text and day_text):
            return None
        if month_text.isdigit():
            month = int(month_text)
        else:
            month = MONTHS[month_text.capitalize()]
        return date(
            year=int(year_text),
            month=month,
            day=int(day_text)
        )


class PubdateParserYearMonth(PubdateDefaults):
    """只有年份、月份"""
    def __call__(self, element):
        year_text = extract_first(element.xpath('./Year/text()'))
        month_text = extract_first(element.xpath('./Month/text()'))
        if not (year_text and month_text):
            return None
        if month_text.isdigit():
            month = int(month_text)
        else:
            month = MONTHS[month_text.capitalize()]
        return date(year=int(year_text), month=month, day=self.default_day)


class PubdateParserYearSeason(PubdateDefaults):
    """只有年份、季节"""
    def __call__(self, element):
        year_text = extract_first(element.xpath('./Year/text()'))
        season_text = extract_first(element.xpath('./Season/text()'))
        if not (year_text and season_text):
            return None
        year = int(year_text)
        month = SEASONS[season_text]
        return date(year=year, month=month, day=self.default_day)


class PubdateParserYearOnly(PubdateDefaults):
    """只有年份"""
    def __call__(self, element):
        year_text = extract_first(element.xpath('./Year/text()'))
        if not year_text:
            return None
        return date(
            year=int(year_text),
            month=self.default_month,
            day=self.default_day
        )


class PubdateParserMedlineDateYearOnly(PubdateDefaults):
    """MedlineDate字段只有年份"""
    pattern = re.compile(r'^(?P<year>\d{4}$)')

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        return date(
            year=year,
            month=self.default_month,
            day=self.default_day
        )


class PubdateParserMedlineDateMonthRange(PubdateDefaults):
    """MedlineDate字段同一年月份有区间"""
    pattern = re.compile(
        r'^(?P<year>\d{4}) (?P<month_text>[a-zA-Z]{3,})(.*)-(.*)[a-zA-Z]{3}$'
    )
    pattern2 = re.compile(
        r'^(?P<year>\d{4}) (?P<month_text>[a-zA-Z]{3,})(.*)/(.*)[a-zA-Z]{3}$'
    )

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            match = self.pattern2.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        month_txt = match.groupdict()['month_text'].capitalize()
        if month_txt in MONTHS.keys():
            month = MONTHS[month_txt]
        else:
            return None
        return date(
            year=year,
            month=month,
            day=self.default_day
        )


class PubdateParserMedlineDateDayRange(PubdateDefaults):
    """MedlineDate字段同一年月份有区间"""
    pattern = re.compile(
        r'^(?P<year>\d{4}) (?P<month_text>[a-zA-Z]{3,}) (?P<day>\d{1,2})-\d{1,2}$'
    )

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        month_txt = match.groupdict()['month_text'].capitalize()
        if month_txt in MONTHS.keys():
            month = MONTHS[month_txt]
        else:
            return None
        day = int(match.groupdict()['day'])
        return date(
            year=year,
            month=month,
            day=day
        )


class PubdateParserMedlineDateFullMonthYear(PubdateDefaults):
    """MedlineDate字段年 月份为全拼"""
    pattern = re.compile(
        r'^(?P<year>\d{4}) (?P<month_text>[a-zA-Z]{3,})$'
    )

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        month_txt = match.groupdict()['month_text'].capitalize()
        if month_txt in MONTHS.keys():
            month = MONTHS[match.groupdict()['month_text'].capitalize()]
        else:
            return None
        return date(
            year=year,
            month=month,
            day=self.default_day
        )


class PubdateParserMedlineDateMonthRangeCrossYear(PubdateDefaults):
    """MedlineDate字段不同一年月份有区间"""
    pattern = re.compile(
        r'^(?P<year>\d{4}) (?P<month_text>[a-zA-Z]{3,})-\d{4} [a-zA-Z]{3}$'
    )
    pattern2 = re.compile(
        r'^(?P<year>\d{4}) (?P<month_text>\d{1,2})-\d{1,2}$'
    )

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            match = self.pattern2.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        month = int(match.groupdict()['month_text'].capitalize())
        return date(
            year=year,
            month=month,
            day=self.default_day
        )


class PubdateParserMedlineDateRangeYear(PubdateDefaults):
    """MedlineDate字段年份区间"""
    pattern = re.compile(
        r'^(?P<year>\d{4})-\d{4}$'
    )

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        return date(
            year=year,
            month=self.default_month,
            day=self.default_day
        )


class PubdateParserMedlineDateMonthDayRange(PubdateDefaults):
    """MedlineDate字段在同一年月份、日期有区间"""
    pattern = re.compile(
        r'^(?P<year>\d{4}) (?P<month_text>[a-zA-Z]{3,}) (?P<day>\d{1,2})-[a-zA-Z]{3} \d{1,2}$'
    )

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        month = MONTHS[match.groupdict()['month_text'].capitalize()]
        day = int(match.groupdict()['day'])
        return date(
            year=year,
            month=month,
            day=day
        )


class PubdateParserMedlineDateMonthDay(PubdateDefaults):
    """MedlineDate字段在同一年月日"""
    pattern = re.compile(
        r'^(?P<year>\d{4}) (?P<month_text>[a-zA-Z]{3,}) (?P<day>\d{1,2})$'
    )

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        month = MONTHS[match.groupdict()['month_text'].capitalize()]
        day = int(match.groupdict()['day'])
        return date(
            year=year,
            month=month,
            day=day
        )


class PubdateParserMedlineDateYearRangeWithSeason(PubdateDefaults):
    """eg, 1976-1977 Winter"""
    pattern = re.compile(
        r'^(?P<year>\d{4})-\d{4} (?P<season_text>[a-zA-Z]+)$'
    )

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        season_text = match.groupdict()['season_text'].capitalize()
        if season_text in SEASONS.keys():
            month = SEASONS[season_text]
        else:
            return None
        return date(
            year=year,
            month=month,
            day=self.default_day
        )


class PubdateParserMedlineDateYearWithSeason(PubdateDefaults):
    """eg, Winter 1976 or 1976 Winter"""
    pattern = re.compile(
        r'^(?P<year>\d{4}) (?P<season_text>[a-zA-Z]+)$'
    )
    pattern2 = re.compile(
        r'^(?P<season_text>[a-zA-Z]+) (?P<year>\d{4})$'
    )
    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            match = self.pattern2.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        season_text = match.groupdict()['season_text'].capitalize()
        if season_text in SEASONS:
            month = SEASONS[season_text]
        else:
            return None
        return date(
            year=year,
            month=month,
            day=self.default_day
        )


class PubdateParserMedlineDateYearSeasonRange(PubdateDefaults):
    """eg, 1977-1978 Fall-Winter"""
    pattern = re.compile(
        r'^(?P<year>\d{4})-\d{4} (?P<season_text>[a-zA-Z]+)-[a-zA-Z]+$'
    )

    def __call__(self, element):
        medline_date_text = extract_first(element.xpath('./MedlineDate/text()'))
        if not medline_date_text:
            return None
        match = self.pattern.search(medline_date_text)
        if not match:
            return None
        year = int(match.groupdict()['year'])
        season_text = match.groupdict()['season_text'].capitalize()
        if season_text in SEASONS:
            month = SEASONS[season_text]
        else:
            return None
        return date(
            year=year,
            month=month,
            day=self.default_day
        )


PUBDATE_PARSERS = [
    PubdateParserYearMonthDay(),
    PubdateParserYearMonth(),
    PubdateParserYearSeason(),
    PubdateParserYearOnly(),
    PubdateParserMedlineDateYearOnly(),
    PubdateParserMedlineDateMonthRange(),
    PubdateParserMedlineDateDayRange(),
    PubdateParserMedlineDateFullMonthYear(),
    PubdateParserMedlineDateMonthRangeCrossYear(),
    PubdateParserMedlineDateRangeYear(),
    PubdateParserMedlineDateMonthDay(),
    PubdateParserMedlineDateMonthDayRange(),
    PubdateParserMedlineDateYearWithSeason(),
    PubdateParserMedlineDateYearRangeWithSeason(),
    PubdateParserMedlineDateYearSeasonRange(),
]


class JournalElementParserMixin(object):
    @staticmethod
    def parse_issn(element):
        return extract_first(element.xpath('./ISSN/text()'))

    @staticmethod
    def parse_issn_type(element):
        issn_element = extract_first(element.xpath('./ISSN'))
        if issn_element is None:
            return None
        return issn_element.get('IssnType')

    @staticmethod
    def parse_title(element):
        return element.xpath('./Title/text()')[0]

    @staticmethod
    def parse_abbr(element):
        return element.xpath('./ISOAbbreviation/text()')[0]


class Journal(JournalElementParserMixin):
    def __init__(self, issn, issn_type, title, abbr):
        self.issn = issn
        self.issn_type = issn_type
        self.title = title
        self.abbr = abbr

    def __repr__(self):
        return self.title

    def to_dict(self):
        return {
            'issn': self.issn,
            'issn_type': self.issn_type,
            'title': self.title,
            'abbr': self.abbr
        }

    @classmethod
    def parse_element(cls, element):
        issn = cls.parse_issn(element)
        issn_type = cls.parse_issn_type(element)
        title = cls.parse_title(element)
        abbr = cls.parse_abbr(element)
        return cls(
            issn=issn, issn_type=issn_type, title=title, abbr=abbr
        )


class Reference(object):
    def __init__(self, citation, ids):
        self.citation = citation
        self.ids = ids

    def __repr__(self):
        return self.citation

    def to_dict(self):
        return {
            'citation': self.citation,
            'ids': [_.to_dict() for _ in self.ids]
        }

    @classmethod
    def parse_element(cls, element):
        """
        parse <Reference></Reference> tag. eg,
        <Reference>
            <Citation>Metabolism. 2009 Jan;58(1):102-8</Citation>
            <ArticleIdList>
                <ArticleId IdType="pubmed">19059537</ArticleId>
            </ArticleIdList>
        </Reference>
        """
        citation = element.xpath('./Citation/text()')[0]
        ids = [
            ArticleId.parse_element(
                article_id_element
            ) for article_id_element in element.xpath(
                './ArticleIdList/ArticleId'
            )
        ]
        return cls(citation=citation, ids=ids)


class AuthorElementParserMixin(object):
    @staticmethod
    def parse_last_name(element):
        return extract_first(element.xpath('./LastName/text()'))

    @staticmethod
    def parse_forename(element):
        return extract_first(element.xpath('./ForeName/text()'))

    @staticmethod
    def parse_initials(element):
        return extract_first(element.xpath('./Initials/text()'))

    @staticmethod
    def parse_affiliation(element):
        return extract_first(element.xpath('./AffiliationInfo/Affiliation/text()'))


class Author(AuthorElementParserMixin):
    def __init__(self, last_name, forename, initials, affiliation):
        """
        Args:
            last_name: 姓
            forename: 名
            initials:
        """
        self.last_name = last_name
        self.forename = forename
        self.initials = initials
        self.affiliation = affiliation

    def __repr__(self):
        return '%s %s %s' % (self.last_name, self.initials, self.forename)

    def to_dict(self):
        return {
            'last_name': self.last_name,
            'forename': self.forename,
            'initials': self.initials,
            'affiliation': self.affiliation
        }

    @classmethod
    def parse_element(cls, element):
        last_name = cls.parse_last_name(element)
        forename = cls.parse_forename(element)
        initials = cls.parse_initials(element)
        affiliation = cls.parse_affiliation(element)
        return cls(
            last_name=last_name,
            forename=forename,
            initials=initials,
            affiliation=affiliation
        )


class ArticleElementParserMixin(object):
    @staticmethod
    def _parse_pmid(element):
        return element.xpath('./MedlineCitation/PMID/text()')[0]

    @staticmethod
    def parse_ids(element):
        return [
            ArticleId.parse_element(
                article_id_element
            ) for article_id_element in element.xpath(
                './PubmedData/ArticleIdList/ArticleId'
            )
        ]

    @staticmethod
    def parse_mesh(element):
        return [
            Mesh.parse_element(mesh_element) for mesh_element in element.xpath(
                './MedlineCitation/MeshHeadingList/MeshHeading/DescriptorName'
            )
        ]
    
    @staticmethod
    def parse_publication(element):
        return [
            Publication.parse_element(pub_element) for pub_element in element.xpath(
                './MedlineCitation/Article/PublicationTypeList/PublicationType'
            )
        ]

    @staticmethod
    def parse_title(element):
        title_element = extract_first(element.xpath(
            './MedlineCitation/Article/ArticleTitle'
        ))
        return get_inner_html(title_element)

    @staticmethod
    def parse_page(element):
        page_element = extract_first(element.xpath(
            './MedlineCitation/Article/Pagination/MedlinePgn'
        ))
        return get_inner_html(page_element)

    @staticmethod
    def parse_language(element):
        page_element = extract_first(element.xpath(
            './MedlineCitation/Article/Language'
        ))
        return get_inner_html(page_element)

    @staticmethod
    def parse_country(element):
        page_element = extract_first(element.xpath(
            './MedlineCitation/MedlineJournalInfo/Country'
        ))
        return get_inner_html(page_element)

    @staticmethod
    def parse_abstract(element):
        paragraphs = []
        for abstract_text_element in element.xpath(
                './MedlineCitation/Article/Abstract/AbstractText'
        ):
            label = abstract_text_element.get('Label', None)
            sub_title = ''
            if label:
                label = label.capitalize()
                sub_title = '%s:' % label
            paragraph = '%s%s' % (sub_title, get_inner_html(abstract_text_element))
            paragraphs.append(paragraph)
        return ''.join(paragraphs)

    @staticmethod
    def parse_keywords(element):
        return element.xpath('./MedlineCitation/KeywordList/Keyword/text()')

    @staticmethod
    def parse_authors(element):
        return [
            Author.parse_element(element) for element in element.xpath(
                './MedlineCitation/Article/AuthorList/Author'
            )
        ]

    @staticmethod
    def parse_journal(element):
        return Journal.parse_element(
            element.xpath('./MedlineCitation/Article/Journal')[0]
        )

    @staticmethod
    def parse_volume(element):
        return extract_first(element.xpath(
            './MedlineCitation/Article/Journal/JournalIssue/Volume/text()'
        ))

    @staticmethod
    def parse_issue(element):
        return extract_first(element.xpath(
            './MedlineCitation/Article/Journal/JournalIssue/Issue/text()'
        ))

    @staticmethod
    def parse_references(element):
        return [
            Reference.parse_element(
                reference_element
            ) for reference_element in element.xpath(
                './PubmedData/ReferenceList/Reference'
            )
        ]

    @staticmethod
    def parse_pubdate(element):
        pubdate_element = element.xpath(
            './MedlineCitation/Article/Journal/JournalIssue/PubDate'
        )[0]
        pubdate = None
        for parser in PUBDATE_PARSERS:
            pubdate = parser(pubdate_element)
            if pubdate:
                break
        if pubdate is None:
            raise PubmedMapperError('日期无法解析，日期格式：%s' % etree.tostring(
                pubdate_element, encoding='utf-8', pretty_print=True
            ))
        return pubdate


class Article(ArticleElementParserMixin):
    def __init__(
            self,
            pmid,
            ids,
            title,
            abstract,
            keywords,
            mesh_headings,
            authors,
            journal,
            publications,
            volume,
            issue,
            references,
            pubdate,
            page,
            language,
            country
    ):
        self.pmid = pmid
        self.ids = ids
        self.title = title
        self.abstract = abstract
        self.keywords = keywords
        self.mesh_headings = mesh_headings
        self.authors = authors
        self.publications = publications
        self.journal = journal
        self.volume = volume
        self.issue = issue
        self.references = references
        self.pubdate = pubdate
        self.page = page
        self.language = language
        self.country = country

    def __repr__(self):
        return self.title

    def to_dict(self):
        return {
            'pmid': self.pmid,
            'ids': [_.to_dict() for _ in self.ids],
            'title': self.title,
            'abstract': self.abstract,
            'keywords': self.keywords,
            'mesh_headings': self.mesh_headings,
            'authors': [author.to_dict() for author in self.authors],
            'publications': self.publications,
            'journal': self.journal.to_dict(),
            'volume': self.volume,
            'issue': self.issue,
            'references': [reference.to_dict() for reference in self.references],
            'pubdate': self.pubdate.strftime('%Y-%m-%d'),
            'page': self.page,
            'language': self.language,
            'country': self.country
        }

    @classmethod
    def parse_element(cls, element):
        pmid = cls._parse_pmid(element)
        ids = cls.parse_ids(element)
        title = cls.parse_title(element)
        abstract = cls.parse_abstract(element)
        keywords = cls.parse_keywords(element)
        mesh_headings = cls.parse_mesh(element)
        authors = cls.parse_authors(element)
        journal = cls.parse_journal(element)
        publications = cls.parse_publication(element)
        volume = cls.parse_volume(element)
        issue = cls.parse_issue(element)
        references = cls.parse_references(element)
        pubdate = cls.parse_pubdate(element)
        page = cls.parse_page(element)
        language = cls.parse_language(element)
        country = cls.parse_country(element)
        return Article(
            pmid=pmid,
            ids=ids,
            title=title,
            abstract=abstract,
            keywords=keywords,
            mesh_headings=mesh_headings,
            authors=authors,
            journal=journal,
            publications=publications,
            volume=volume,
            issue=issue,
            references=references,
            pubdate=pubdate,
            page=page,
            language=language,
            country=country
        )

    @classmethod
    def parse_pmid(cls, pmid):
        url = ('https://eutils.ncbi.nlm.nih.gov'
               '/entrez/eutils/efetch.fcgi?'
               'db=pubmed&id=%s&retmode=xml') % pmid
        try:
            handle = urlopen(url, timeout=60)
        except (URLError, HTTPError) as e:
            print('cannot download %s', pmid)
            print(e)
            return None
        root = etree.parse(handle)
        try:
            element = root.xpath('/PubmedArticleSet/PubmedArticle')[0]
        except Exception as e:
            return None
        return cls.parse_element(element)



# -*- encoding:utf-8 -*-
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = '0.2.1'

setuptools.setup(
    name="pubmed_download",
    version=version,
    author="lin.li",
    author_email="lilinxr@gmail.com",
    description="download pdf by pmid",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/decouples/pubmed_download",
    packages=setuptools.find_packages(),
    install_requires=['setuptools',
                      'pandas',
                      'lxml',
                      'requests',
                      'PyPDF2',
                      'bs4'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# -*- encoding:utf-8 -*-
"""
@author：lin.li
@filename：pubmed_download.py
@time：2022/8/6  22:41
@file description:pubmed_download
"""
import traceback
import pandas as pd
import requests
import os
import PyPDF2
import urllib.request as ub
from bs4 import BeautifulSoup
from .pubmed_mapper import Article
import logging


logger = logging.getLogger('pubmed')
logger.propagate = False
logger.setLevel(logging.DEBUG)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
formatter = logging.Formatter('[{asctime}] [{name:^5}] [{levelname:^7}] [{lineno:^4}] [{message}]',
                              datefmt='%Y-%m-%d %H:%M:%S', style='{',)
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


class DownloadPdf:
    """
    给定pmid列表和下载路径，下载对应的pdf文件

    失败的文件写在csv文件中
    """

    def __init__(self, store_path, pmid_list):
        self.store_path = store_path
        self.logger = logger
        self.pmid_list = pmid_list
        self.failed_list = list()

    def download(self, pmid, url):
        """真实下载任务函数

        根据所给pdf下载，根据pmid命名pdf，保存到指定文件夹

        如果url请求失败或者下载的不是pdf删除下载的文件，并返回False
        """
        headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:86.0) Gecko/20100101 Firefox/86.0',
                   'Upgrade-Insecure-Requests': '1',
                   'Cookie': '__cfduid=d9b6123514d7ca2db292dca147121a9041616482481'
                   }
        try:
            response = requests.get(url, stream=True, headers=headers, timeout=60)
            if response.status_code != 200:
                return False
            filename = self.store_path + '/' + os.path.basename(str(pmid)) + '.pdf'
            self.logger.info(f"{pmid} {url}")
            with open(filename, 'wb') as file:
                for data in response.iter_content():
                    file.write(data)
            try:
                with open(filename, 'rb') as f:
                    PyPDF2.PdfFileReader(f)
            except ValueError:
                return True
            except Exception as e:
                self.logger.error(f"{pmid}, download pdf error, {e}")
                os.remove(filename)
                return False
        except Exception as e:
            self.logger.error(f"{pmid}, cannot download, {e}")
            return False
        return True

    def check_downloaded(self, pmid):
        """check pdf exists"""
        pdf_path = os.path.join(self.store_path, f"{pmid}.pdf")
        if os.path.exists(pdf_path) and os.path.isfile(pdf_path):
            return True
        return False

    def run(self):
        """这里做主要的下载工作"""
        self.logger.info("start download.....")
        for pmid in self.pmid_list:
            # time.sleep(3)
            self.logger.info(f"start download {pmid}, {self.pmid_list.index(pmid)}/{len(self.pmid_list)}")
            # 该批次下载过的不再下载
            if self.check_downloaded(pmid):
                self.logger.info(f"file already exist:{pmid}")
                self.failed_list.append(pmid)
                continue
            try:
                article = Article.parse_pmid(pmid)
            except Exception as e:
                self.logger.error(f"parse pmid failed: {pmid}, reason: {traceback.print_exc()}")
                self.failed_list.append(pmid)
                continue
            ids_dict = dict()
            [ids_dict.update({k.id_type: k.id_value}) for k in article.ids]
            doi = ids_dict.get('doi', None)
            pii = ids_dict.get('pii', None)
            pmc = ids_dict.get('pmc', None)
            if doi and self.search_by_doi(pmid, doi.strip()):
                continue
            if pii and self.search_by_pii(pmid, pii.strip()):
                continue
            if pmc and self.search_by_pmc(pmid, pmc.strip()):
                continue
            self.logger.info(f"download failed: {pmid}")
            self.failed_list.append(pmid)
        result_csv = os.path.join(os.path.dirname(self.store_path), "failed_list.csv")
        df = pd.DataFrame(self.failed_list, columns=['pmid'])
        df.to_csv(result_csv, sep='\t', index=False)
        self.logger.info(f"download file done:{len(self.pmid_list) - len(self.failed_list)}/{len(self.pmid_list)}")
        return list(set(self.pmid_list) - set(self.failed_list))

    def search_by_doi(self, pmid, doi):
        """这里使用文章的doi进行文章下载"""
        self.logger.info(f"search by doi: {pmid}, {doi}")
        download_status = False
        temp_url = 'https://sci-hub.ren/' + doi
        opener = ub.build_opener()
        opener.addheaders = [
            {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:86.0) Gecko/20100101 Firefox/86.0',
             'Upgrade-Insecure-Requests': 1}]
        ub.install_opener(opener)
        try:
            content = ub.urlopen(temp_url).read().decode('utf-8')
            soup = BeautifulSoup(content, 'html.parser')
            pdf_element = soup.find('iframe', id='pdf')
            if pdf_element is not None:
                pdf_url = pdf_element.attrs['src']
                if 'https:' not in pdf_url:
                    pdf_url = 'https:' + pdf_url
                self.logger.info(f"search sci_hub by doi:{pdf_url}")
                download_status = self.download(pmid, pdf_url)
        except Exception as e:
            self.logger.error(f"search sci-hub by doi error: {traceback.print_exc()}")

        candidate_url_list = [
            f"https://sci-hub.wf/{doi}",
            f'https://sci/bban/top/{doi}.pdf#view=FitH',
            f"https://www.spandidos-publications.com/{doi}/download",
            f"https://onlinelibrary.wiley.com/doi/pdfdirect/{doi}",
            f"https://jeccr.biomedcentral.com/track/pdf/{doi}",
            f"https://www.frontiersin.org/articles/{doi}/pdf",
            f"https://journals.plos.org/plosone/article/file?id={doi}&type=printable",
            f"https://onlinelibrary.wiley.com/doi/pdfdirect/{doi}",
            f"https://www.tandfonline.com/doi/pdf/{doi}?needAccess=true",
            f"https://immunityageing.biomedcentral.com/track/pdf/{doi}",
            f"https://pubs.acs.org/doi/pdf/{doi}",
            f"https://journals.sagepub.com/doi/pdf/{doi}",
            f"https://www.futuremedicine.com/doi/epub/{doi}"
        ]
        for url in candidate_url_list:
            if not download_status:
                self.logger.info(f"try by doi: {url}")
                self.download(pmid, url.strip())
            else:
                break
        self.logger.info(f"download by doi:{download_status}")
        return download_status

    def search_by_pii(self, pmid, pii):
        """这里使用pii下载文章"""
        self.logger.info(f"search by pii, {pmid}, {pii}...")
        download_status = False
        candidate_url_list = [
            f'https://www.jto.org/action/showPdf?pii={pii}',
            f"https://www.jto.org/article/{pii}/pdf",
            f"https://www.oncotarget.com/article/{pii}/pdf/"
        ]
        for url in candidate_url_list:
            if not download_status:
                self.logger.info(f'try by pii: {url}')
                download_status = self.download(pmid, url)
            else:
                break
        self.logger.info(f"download by pii: {download_status}")
        return download_status

    def search_by_pmc(self, pmid, pmc):
        """根据pmc尝试去对应网站下载pdf"""
        download_status = False
        try:
            temp_url = f'https://www.ncbi.nlm.nih.gov/pmc/articles/{pmc}'
            content = ub.urlopen(temp_url).read().decode('utf-8')
            soup = BeautifulSoup(content, 'html.parser')
            pdf_url = 'https://www.ncbi.nlm.nih.gov' + \
                      soup.find('link', type='application/pdf').attrs['href']
            self.logger.info(f"try by pmc {pdf_url}")
            download_status = self.download(pmid, pdf_url)
        except Exception as e:
            self.logger.error(f"{pmid}, concat pmc-url error-->, {e}")
        candidate_url_list = [
            f"http://europepmc.org/articles/{pmc.lower()}?pdf=render"
        ]
        for url in candidate_url_list:
            if not download_status:
                self.logger.info(f"try by pmc: {url}")
                download_status = self.download(pmid, url)
            else:
                break
        self.logger.info(f"download by pmc: {download_status}")
        return download_status
